/*
 * Copyright © 2011 Linaro Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Linaro Ltd. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Linaro Ltd. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * LINARO LTD. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL LINARO LTD. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Travis Watkins <travis.watkins@linaro.org>
 */

#define GL_GLEXT_PROTOTYPES

#include <iostream>
#include <fstream>
#include <opengl/program.h>

class PrivateProgram
{
    public:
	GLuint program;
	bool valid;
};


void printShaderInfoLog (GLuint shader)
{
    GLint   length = 0;
    GLint   chars  = 0;
    GLchar *infoLog;

    glGetShaderiv (shader, GL_INFO_LOG_LENGTH, &length);

    if (length > 0)
    {
	infoLog = new GLchar[length];
	glGetShaderInfoLog (shader, length, &chars, infoLog);
	std::cout << infoLog << std::endl;
	delete[] infoLog;
    }
}

void printProgramInfoLog(GLuint program)
{
    GLint   length = 0;
    GLint   chars  = 0;
    GLchar *infoLog;

    glGetProgramiv (program, GL_INFO_LOG_LENGTH, &length);

    if (length > 0)
    {
	infoLog = new GLchar[length];
	glGetProgramInfoLog (program, length, &chars, infoLog);
	std::cout << infoLog << std::endl;
	delete[] infoLog;
    }
}

static bool compileShader (GLuint *shader, GLenum type, CompString &path)
{
    std::ifstream in;
    const GLchar *data;
    GLint         status;

    if (path[0] != '/')
    {
	std::string temp = METADATADIR;
	temp.append ("/");
	temp.append (path);
	path = temp;
    }

    in.open (path.c_str (), std::ios::in);

    std::string temp((std::istreambuf_iterator<char>(in)), std::istreambuf_iterator<char>());
    in.close ();

    data = (GLchar *)temp.c_str ();

    *shader = glCreateShader (type);
    glShaderSource (*shader, 1, &data, NULL);
    glCompileShader (*shader);

    glGetShaderiv (*shader, GL_COMPILE_STATUS, &status);
    return (status == GL_TRUE);
}

GLProgram::GLProgram (CompString &vertexFile, CompString &fragmentFile) :
    priv (new PrivateProgram ())
{
    GLuint vertex, fragment;
    GLint status;

    priv->valid = false;
    priv->program = glCreateProgram ();

    if (!compileShader (&vertex, GL_VERTEX_SHADER, vertexFile))
    {
	printShaderInfoLog (vertex);
	return;
    }

    if (!compileShader (&fragment, GL_FRAGMENT_SHADER, fragmentFile))
    {
	printShaderInfoLog (fragment);
	return;
    }

    glAttachShader (priv->program, vertex);
    glAttachShader (priv->program, fragment);

    glLinkProgram (priv->program);
    glValidateProgram (priv->program);

    glGetProgramiv (priv->program, GL_LINK_STATUS, &status);
    if (status == GL_FALSE)
    {
	printProgramInfoLog (priv->program);
	return;
    }

    glDeleteShader (vertex);
    glDeleteShader (fragment);

    priv->valid = true;
}

GLProgram::~GLProgram ()
{
    glDeleteProgram (priv->program);
    delete priv;
}

bool GLProgram::valid ()
{
    return priv->valid;
}

void GLProgram::bind ()
{
    glUseProgram (priv->program);
}

void GLProgram::unbind ()
{
    glUseProgram (0);
}

bool GLProgram::setUniform (const char *name, GLfloat value)
{
    GLint location = glGetUniformLocation (priv->program, name);
    if (location == -1)
	return false;

    glUniform1f (location, value);
    return true;
}

bool GLProgram::setUniform (const char *name, GLint value)
{
    GLint location = glGetUniformLocation (priv->program, name);
    if (location == -1)
	return false;

    glUniform1i (location, value);
    return true;
}

bool GLProgram::setUniform (const char *name, const GLMatrix &value)
{
    GLint location = glGetUniformLocation (priv->program, name);
    if (location == -1)
	return false;

    glUniformMatrix4fv (location, 1, GL_FALSE, value.getMatrix ());
    return true;
}

bool GLProgram::setUniform2f (const char *name,
                              GLfloat x,
                              GLfloat y)
{
    GLint location = glGetUniformLocation (priv->program, name);
    if (location == -1)
	return false;

    glUniform2f (location, x, y);
    return true;
}

bool GLProgram::setUniform3f (const char *name,
                              GLfloat x,
                              GLfloat y,
                              GLfloat z)
{
    GLint location = glGetUniformLocation (priv->program, name);
    if (location == -1)
	return false;

    glUniform3f (location, x, y, z);
    return true;
}

bool GLProgram::setUniform4f (const char *name,
                              GLfloat x,
                              GLfloat y,
                              GLfloat z,
                              GLfloat w)
{
    GLint location = glGetUniformLocation (priv->program, name);
    if (location == -1)
	return false;

    glUniform4f (location, x, y, z, w);
    return true;
}

GLuint GLProgram::attributeLocation (const char *name)
{
    return glGetAttribLocation (priv->program, name);
}

