/*
 * Copyright © 2011 Linaro Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of
 * Linaro Ltd. not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Linaro Ltd. makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * LINARO LTD. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL LINARO LTD. BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Travis Watkins <travis.watkins@linaro.org>
 */

#define GL_GLEXT_PROTOTYPES

#include <vector>
#include <iostream>

#ifdef USE_GLES
#include <GLES2/gl2.h>
#else
#include <GL/gl.h>
#include <GL/glext.h>
#endif

#include <opengl/vertexbuffer.h>

#include "privates.h"

GLVertexBuffer *PrivateVertexBuffer::streamingBuffer = NULL;
GLProgram *PrivateVertexBuffer::genericProgram = NULL;

GLVertexBuffer::GLVertexBuffer () :
    priv (new PrivateVertexBuffer ())
{
    priv->usage = GL_STATIC_DRAW;
}

GLVertexBuffer::GLVertexBuffer (GLenum usage) :
    priv (new PrivateVertexBuffer ())
{
    priv->usage = usage;
}

GLVertexBuffer::~GLVertexBuffer ()
{
    delete priv;
}

GLVertexBuffer *GLVertexBuffer::streamingBuffer ()
{
    if (PrivateVertexBuffer::streamingBuffer == NULL)
	PrivateVertexBuffer::streamingBuffer = new GLVertexBuffer
	                                                      (GL_STREAM_DRAW);
    return PrivateVertexBuffer::streamingBuffer;
}

void GLVertexBuffer::begin (GLenum primitiveType)
{
    priv->primitiveType = primitiveType;

    priv->vertexData.clear ();
    priv->normalData.clear ();
    priv->colorData.clear ();
    priv->textureData.clear ();
}

void GLVertexBuffer::begin ()
{
    begin (GL_TRIANGLES);
}

int GLVertexBuffer::end ()
{
    if (!GL::vbo)
	return 0;

    if (!priv->vertexData.size ())
	return -1;

    GL::bindBuffer (GL_ARRAY_BUFFER, priv->vertexBuffer);
    GL::bufferData (GL_ARRAY_BUFFER,
                    sizeof(GLfloat) * priv->vertexData.size () * 3,
                    &priv->vertexData[0], priv->usage);

    if (priv->normalData.size ())
    {
	GL::bindBuffer (GL_ARRAY_BUFFER, priv->normalBuffer);
	GL::bufferData (GL_ARRAY_BUFFER,
	                sizeof(GLfloat) * priv->normalData.size () * 3,
	                &priv->normalData[0], priv->usage);
    }

    if (priv->colorData.size ())
    {
	GL::bindBuffer (GL_ARRAY_BUFFER, priv->colorBuffer);
	GL::bufferData (GL_ARRAY_BUFFER,
	                sizeof(GLfloat) * priv->colorData.size () * 4,
	                &priv->colorData[0], priv->usage);
    }

    if (priv->textureData.size ())
    {
	for (unsigned int i = 0; i < priv->textureData.size (); i++)
	{
	    GL::bindBuffer (GL_ARRAY_BUFFER, priv->textureBuffers[i]);
	    GL::bufferData (GL_ARRAY_BUFFER,
	                    sizeof(GLfloat) * priv->textureData[i].size () * 2,
	                    &priv->textureData[i][0], priv->usage);
	}
    }

    GL::bindBuffer (GL_ARRAY_BUFFER, 0);

    return 0;
}

void GLVertexBuffer::addVertices (GLuint nVertices, GLfloat *vertices)
{
    priv->vertexData.reserve (priv->vertexData.size () + (nVertices * 3));

    for (GLuint i = 0; i < nVertices * 3; i++)
    {
	priv->vertexData.push_back (vertices[i]);
    }
}

void GLVertexBuffer::addNormals (GLuint nNormals, GLfloat *normals)
{
    priv->normalData.reserve (priv->normalData.size () + (nNormals * 3));

    for (GLuint i = 0; i < nNormals * 3; i++)
    {
	priv->normalData.push_back (normals[i]);
    }
}

void GLVertexBuffer::addColors (GLuint nColors, GLushort *colors)
{
    priv->colorData.reserve (priv->colorData.size () + (nColors * 4));

    for (GLuint i = 0; i < nColors * 4; i++)
    {
	priv->colorData.push_back ((colors[i] & 0xFF) / 255.0f);
    }
}

void GLVertexBuffer::addTexCoords (GLuint texture,
                                   GLuint nTexcoords,
                                   GLfloat *texcoords)
{
    //four textures max (zero indexed)
    if (texture > 3)
	return;

    while (texture >= priv->textureData.size ())
    {
	std::vector<GLfloat> temp;
	priv->textureData.push_back (temp);
    }

    priv->textureData[texture].reserve (priv->textureData.size () +
                                        (nTexcoords * 2));

    for (GLuint i = 0; i < nTexcoords * 2; i++)
    {
	priv->textureData[texture].push_back (texcoords[i]);
    }
}

void GLVertexBuffer::useProgram (GLProgram *program)
{
    priv->customProgram = program;
}

int GLVertexBuffer::render (const GLMatrix &modelview)
{
    const GLWindowPaintAttrib attrib = { OPAQUE, BRIGHT, COLOR, 0, 0, 0, 0 };

    return render (modelview, attrib);
}

int GLVertexBuffer::render (const GLMatrix            &modelview,
                            const GLWindowPaintAttrib &attrib)
{
    GLScreen *gScreen = GLScreen::get (screen);
    GLMatrix *projection = gScreen->projectionMatrix ();

    return render (*projection, modelview, attrib);
}

int GLVertexBuffer::render (const GLMatrix            &projection,
                            const GLMatrix            &modelview,
                            const GLWindowPaintAttrib &attrib)
{
    if (!priv->vertexData.size ())
	return -1;

    glEnable (GL_BLEND);

    if (GL::vbo)
	return priv->render (projection, modelview, attrib);
    else
	return priv->legacyRender (projection, modelview, attrib);

    glDisable (GL_BLEND);
}

PrivateVertexBuffer::PrivateVertexBuffer () :
    customProgram (NULL)
{
    if (!GL::vbo)
	return;

    GL::genBuffers (1, &vertexBuffer);
    GL::genBuffers (1, &normalBuffer);
    GL::genBuffers (1, &colorBuffer);

    textureBuffers.reserve (4);
    GL::genBuffers (4, &textureBuffers[0]);

    if (PrivateVertexBuffer::genericProgram == NULL)
    {
	CompString vertex("generic_vertex.glsl");
	CompString fragment("generic_fragment.glsl");
	PrivateVertexBuffer::genericProgram = new GLProgram (vertex, fragment);
    }
}

PrivateVertexBuffer::~PrivateVertexBuffer ()
{
    if (!GL::vbo)
	return;

    GL::deleteBuffers (1, &vertexBuffer);
    GL::deleteBuffers (1, &normalBuffer);
    GL::deleteBuffers (1, &colorBuffer);
    GL::deleteBuffers (4, &textureBuffers[0]);
}

int PrivateVertexBuffer::render (const GLMatrix            &projection,
                                 const GLMatrix            &modelview,
                                 const GLWindowPaintAttrib &attrib)
{
    GLfloat params[4] = {0, 0, 0, 0};
    GLfloat attribs[3] = {1, 1, 1};
    GLint index = 0;
    GLProgram *program;

    if (customProgram == NULL)
	program = genericProgram;
    else
	program = customProgram;

    program->bind ();
    if (!program->valid ())
    {
	std::cerr << "invalid shader program!" << std::endl;
	return -1;
    }

    program->setUniform ("projection", projection);
    program->setUniform ("modelview", modelview);

    index = program->attributeLocation ("position");
    glEnableVertexAttribArray (index);
    GL::bindBuffer (GL_ARRAY_BUFFER, vertexBuffer);
    glVertexAttribPointer (index, 3, GL_FLOAT, GL_FALSE, 0, 0);

    //use default normal
    if (normalData.size () == 0)
    {
	program->setUniform3f ("singleNormal", 0.0f, 0.0f, -1.0f);
    }
    // special case a single normal and apply it to the entire operation
    else if (normalData.size () == 3)
    {
	program->setUniform3f ("singleNormal",
	                       normalData[0], normalData[1], normalData[2]);
    }
    else if (normalData.size () > 3)
    {
	index = program->attributeLocation ("normal");
	glEnableVertexAttribArray (index);
	GL::bindBuffer (GL_ARRAY_BUFFER, normalBuffer);
	glVertexAttribPointer (index, 3, GL_FLOAT, GL_FALSE, 0, 0);
    }

    // special case a single color and apply it to the entire operation
    if (colorData.size () == 4)
    {
	program->setUniform4f ("singleColor", colorData[0],
	                       colorData[1], colorData[2], colorData[3]);
	params[0] = 1.0;
    }
    else if (colorData.size () > 4)
    {
	index = program->attributeLocation ("color");
	glEnableVertexAttribArray (index);
	GL::bindBuffer (GL_ARRAY_BUFFER, colorBuffer);
	glVertexAttribPointer (index, 4, GL_FLOAT, GL_FALSE, 0, 0);
	params[1] = 1.0;
    }

    //divide number of textures by 4 to get 0-1 range
    params[2] = textureData.size () / 4.0f;

    for (int i = textureData.size () - 1; i >= 0; i--)
    {
	char name[10];

	snprintf (name, 10, "texCoord%d", i);
	index = program->attributeLocation (name);

	glEnableVertexAttribArray (index);
	GL::bindBuffer (GL_ARRAY_BUFFER, textureBuffers[i]);
	glVertexAttribPointer (index, 2, GL_FLOAT, GL_FALSE, 0, 0);

	snprintf (name, 9, "texture%d", i);
	program->setUniform (name, i);
    }

    //convert paint attribs to 0-1 range
    attribs[0] = (attrib.opacity & 0xFF) / 255.0f;
    attribs[1] = (attrib.brightness & 0xFF) / 255.0f;
    attribs[2] = (attrib.saturation & 0xFF) / 255.0f;
    program->setUniform3f ("paintAttrib", attribs[0], attribs[1], attribs[2]);

    program->setUniform4f ("params", params[0], params[1], params[2], params[3]);
    glDrawArrays (primitiveType, 0, vertexData.size () / 3);

    GL::bindBuffer (GL_ARRAY_BUFFER, 0);
    program->unbind ();

    return 0;
}

int PrivateVertexBuffer::legacyRender (const GLMatrix            &projection,
                                       const GLMatrix            &modelview,
                                       const GLWindowPaintAttrib &attrib)
{
    #ifndef USE_GLES
    glMatrixMode (GL_PROJECTION);
    glPushMatrix ();
    glLoadMatrixf (projection.getMatrix ());

    glMatrixMode (GL_MODELVIEW);
    glPushMatrix ();
    glLoadMatrixf (modelview.getMatrix ());

    glEnableClientState (GL_VERTEX_ARRAY);
    glVertexPointer (3, GL_FLOAT, 0, &vertexData[0]);

    //use default normal
    if (normalData.size () == 0)
    {
	glNormal3f (0.0f, 0.0f, -1.0f);
    }
    // special case a single normal and apply it to the entire operation
    else if (normalData.size () == 3)
    {
	glNormal3fv (&normalData[0]);
    }
    else if (normalData.size () > 3)
    {
	glEnableClientState (GL_NORMAL_ARRAY);
	glNormalPointer (GL_FLOAT, 0, &normalData[0]);
    }

    // special case a single color and apply it to the entire operation
    if (colorData.size () == 4)
    {
	glColor4fv (&colorData[0]);
    }
    else if (colorData.size () > 4)
    {
	glEnableClientState (GL_COLOR_ARRAY);
	glColorPointer (4, GL_FLOAT, 0, &colorData[0]);
    }

    for (int i = textureData.size () - 1; i >= 0; i--)
    {
	GL::clientActiveTexture (GL_TEXTURE0_ARB + i);
	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glTexCoordPointer (2, GL_FLOAT, 0, &textureData[i][0]);
    }

    glDrawArrays (primitiveType, 0, vertexData.size () / 3);

    glDisableClientState (GL_VERTEX_ARRAY);
    glDisableClientState (GL_NORMAL_ARRAY);
    glDisableClientState (GL_COLOR_ARRAY);

    for (int i = textureData.size (); i > 0; i--)
    {
	GL::clientActiveTexture (GL_TEXTURE0_ARB + i);
	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
    }

    GL::clientActiveTexture (GL_TEXTURE0_ARB);

    glMatrixMode (GL_PROJECTION);
    glPopMatrix ();

    glMatrixMode (GL_MODELVIEW);
    glPopMatrix ();
    #endif

    return 0;
}

